{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "biblical-burke",
   "metadata": {},
   "source": [
    "# Project 1: Group Restaurants Choice\n",
    "\n",
    "By: Jacob Andreesen, Jeff Chen, Miao Xu, Yiyi Wang"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "copyrighted-exclusion",
   "metadata": {},
   "source": [
    "Finding an ideal restaurants for students with a group of friends is always a struggle for newcomers, who are looking for places new and excited to go. Inspired by such a challenge, our group aim to build a personalized restaurant recommender system prototype that serve a small group of people to meet their requirements and close to their taste."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "id": "dirty-possibility",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 1.46 s (started: 2021-03-09 15:47:27 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 1;\n",
       "                var nbb_unformatted_code = \"%load_ext autotime\\n%load_ext nb_black\\n%matplotlib inline\\n%config InlineBackend.figure_format = 'retina'\\n\\n# For scraping\\nimport time\\nimport urllib.request, json\\nfrom flatten_dict import flatten\\nimport requests\\nimport copyheaders\\nfrom bs4 import BeautifulSoup\\n\\n# General tools\\nimport regex as re\\nimport matplotlib.pyplot as plt\\nimport numpy as np\\nimport pandas as pd\\nfrom tqdm import tqdm\\nimport geopandas\\nfrom geopy import Nominatim\";\n",
       "                var nbb_formatted_code = \"%load_ext autotime\\n%load_ext nb_black\\n%matplotlib inline\\n%config InlineBackend.figure_format = 'retina'\\n\\n# For scraping\\nimport time\\nimport urllib.request, json\\nfrom flatten_dict import flatten\\nimport requests\\nimport copyheaders\\nfrom bs4 import BeautifulSoup\\n\\n# General tools\\nimport regex as re\\nimport matplotlib.pyplot as plt\\nimport numpy as np\\nimport pandas as pd\\nfrom tqdm import tqdm\\nimport geopandas\\nfrom geopy import Nominatim\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "%load_ext autotime\n",
    "%load_ext nb_black\n",
    "%matplotlib inline\n",
    "%config InlineBackend.figure_format = 'retina'\n",
    "\n",
    "# For scraping\n",
    "import time\n",
    "import urllib.request, json\n",
    "from flatten_dict import flatten\n",
    "import requests\n",
    "import copyheaders\n",
    "from bs4 import BeautifulSoup\n",
    "\n",
    "# General tools\n",
    "import regex as re\n",
    "import matplotlib.pyplot as plt\n",
    "import numpy as np\n",
    "import pandas as pd\n",
    "from tqdm import tqdm\n",
    "import geopandas\n",
    "from geopy import Nominatim"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "available-cancer",
   "metadata": {},
   "source": [
    "---\n",
    "<a id=1></a>\n",
    "## Web Scraping Yelp!\n",
    "\n",
    "We will be simply grabbing data from `https://www.yelp.com/search/snippet` api endpoint instead of actually web scraping from the Yelp! website. We attempted to scrape it through the website but it was hard to select specific elements that we required and some of them are only revealed through button clicks, meaning we'd have to use a browser automation software like selenium to simulate clicks and grab data from the html afterwards, a little too much unecessary work."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "id": "convenient-racing",
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      " 80%|████████  | 20/25 [00:35<00:09,  1.87s/it]"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Number of attempts to get data for 200: 10\n",
      "Number of attempts to get data for 200: 20\n",
      "Number of attempts to get data for 200: 30\n",
      "Number of attempts to get data for 200: 40\n",
      "Number of attempts to get data for 200: 50\n",
      "Number of attempts to get data for 200: 60\n",
      "Number of attempts to get data for 200: 70\n",
      "Number of attempts to get data for 200: 80\n",
      "Number of attempts to get data for 200: 90\n",
      "Number of attempts to get data for 200: 100\n",
      "Number of attempts to get data for 200: 110\n",
      "Number of attempts to get data for 200: 120\n",
      "Number of attempts to get data for 200: 130\n",
      "Number of attempts to get data for 200: 140\n",
      "Number of attempts to get data for 200: 150\n",
      "Number of attempts to get data for 200: 160\n",
      "Number of attempts to get data for 200: 170\n",
      "Number of attempts to get data for 200: 180\n",
      "Number of attempts to get data for 200: 190\n",
      "Number of attempts to get data for 200: 200\n",
      "Number of attempts to get data for 200: 210\n",
      "Number of attempts to get data for 200: 220\n",
      "Number of attempts to get data for 200: 230\n",
      "Number of attempts to get data for 200: 240\n",
      "Number of attempts to get data for 200: 250\n",
      "Number of attempts to get data for 200: 260\n",
      "Number of attempts to get data for 200: 270\n",
      "Number of attempts to get data for 200: 280\n",
      "Number of attempts to get data for 200: 290\n",
      "Number of attempts to get data for 200: 300\n",
      "Number of attempts to get data for 200: 310\n",
      "Number of attempts to get data for 200: 320\n",
      "Number of attempts to get data for 200: 330\n",
      "Number of attempts to get data for 200: 340\n",
      "Number of attempts to get data for 200: 350\n",
      "Number of attempts to get data for 200: 360\n",
      "Number of attempts to get data for 200: 370\n"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      " 96%|█████████▌| 24/25 [04:38<00:26, 26.14s/it]"
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Number of attempts to get data for 240: 10\n",
      "Number of attempts to get data for 240: 20\n",
      "Number of attempts to get data for 240: 30\n",
      "Number of attempts to get data for 240: 40\n",
      "Number of attempts to get data for 240: 50\n",
      "Number of attempts to get data for 240: 60\n",
      "Number of attempts to get data for 240: 70\n",
      "Number of attempts to get data for 240: 80\n",
      "Number of attempts to get data for 240: 90\n",
      "Number of attempts to get data for 240: 100\n",
      "Number of attempts to get data for 240: 110\n",
      "Number of attempts to get data for 240: 120\n",
      "Number of attempts to get data for 240: 130\n",
      "Number of attempts to get data for 240: 140\n",
      "Number of attempts to get data for 240: 150\n",
      "Number of attempts to get data for 240: 160\n",
      "Number of attempts to get data for 240: 170\n",
      "Number of attempts to get data for 240: 180\n",
      "Number of attempts to get data for 240: 190\n",
      "Number of attempts to get data for 240: 200\n",
      "Number of attempts to get data for 240: 210\n",
      "Number of attempts to get data for 240: 220\n",
      "Number of attempts to get data for 240: 230\n",
      "Number of attempts to get data for 240: 240\n",
      "Number of attempts to get data for 240: 250\n",
      "Number of attempts to get data for 240: 260\n",
      "Number of attempts to get data for 240: 270\n",
      "Number of attempts to get data for 240: 280\n",
      "Number of attempts to get data for 240: 290\n",
      "Number of attempts to get data for 240: 300\n",
      "Number of attempts to get data for 240: 310\n",
      "Number of attempts to get data for 240: 320\n",
      "Number of attempts to get data for 240: 330\n",
      "Number of attempts to get data for 240: 340\n",
      "Number of attempts to get data for 240: 350\n",
      "Number of attempts to get data for 240: 360\n",
      "Number of attempts to get data for 240: 370\n",
      "Number of attempts to get data for 240: 380\n",
      "Number of attempts to get data for 240: 390\n",
      "Number of attempts to get data for 240: 400\n",
      "Number of attempts to get data for 240: 410\n",
      "Number of attempts to get data for 240: 420\n",
      "Number of attempts to get data for 240: 430\n",
      "Number of attempts to get data for 240: 440\n",
      "Number of attempts to get data for 240: 450\n",
      "Number of attempts to get data for 240: 460\n",
      "Number of attempts to get data for 240: 470\n",
      "Number of attempts to get data for 240: 480\n",
      "Number of attempts to get data for 240: 490\n"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "100%|██████████| 25/25 [09:52<00:00, 23.68s/it] "
     ]
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Number of attempts to get data for 240: 500\n",
      "time: 9min 51s (started: 2021-03-09 15:47:29 +08:00)\n"
     ]
    },
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 2;\n",
       "                var nbb_unformatted_code = \"headers_str = b\\\"\\\"\\\"\\n    cache-control: max-age=0, must-revalidate, no-cache, no-store, private\\n    cache-control: no-transform\\n    cf-cache-status: DYNAMIC\\n    cf-ray: 58b26184fbd76c86-SJC\\n    cf-request-id: 02635b471c00006c86b019b200000001\\n    content-encoding: gzip\\n    content-security-policy: report-uri https://www.yelp.com/csp_block?id=bf59639897830a99&page=enforced_by_default_directives&policy_hash=7b6f2d6630868fdb2698dac44731677c&site=www&timestamp=1588093661; object-src 'self'; base-uri 'self' https://*.yelpcdn.com https://*.adsrvr.org https://6372968.fls.doubleclick.net; font-src data: 'self' https://*.yelp.com https://*.yelpcdn.com https://fonts.gstatic.com https://connect.facebook.net https://cdnjs.cloudflare.com https://apis.google.com https://www.google-analytics.com https://use.typekit.net https://player.ooyala.com https://use.fontawesome.com https://maxcdn.bootstrapcdn.com https://fonts.googleapis.com\\n    content-security-policy-report-only: report-uri https://www.yelp.com/csp_report_only?id=bf59639897830a99&page=csp_report_frame_directives%2Cfull_site_ssl_csp_report_directives&policy_hash=9dd00a1a6fbb402584b7ce0c1fdb4d14&site=www&timestamp=1588093661; frame-ancestors 'self' https://*.yelp.com; default-src https:; img-src https: data: https://*.adsrvr.org; script-src https: data: 'unsafe-inline' 'unsafe-eval' blob:; style-src https: 'unsafe-inline' data:; connect-src https:; font-src data: 'self' https://*.yelp.com https://*.yelpcdn.com https://fonts.gstatic.com https://connect.facebook.net https://cdnjs.cloudflare.com https://apis.google.com https://www.google-analytics.com https://use.typekit.net https://player.ooyala.com https://use.fontawesome.com https://maxcdn.bootstrapcdn.com https://fonts.googleapis.com; frame-src https: yelp-webview://* yelp://* data:; child-src https: yelp-webview://* yelp://*; media-src https:; object-src 'self'; base-uri 'self' https://*.yelpcdn.com https://*.adsrvr.org https://6372968.fls.doubleclick.net; form-action https: 'self'\\n    content-type: application/json; charset=utf-8\\n    date: Tue, 28 Apr 2020 17:07:42 GMT\\n    expect-ct: max-age=604800, report-uri=\\\"https://report-uri.cloudflare.com/cdn-cgi/beacon/expect-ct\\\"\\n    expires: Tue, 28 Apr 2020 17:07:41 GMT\\n    pragma: no-cache\\n    referrer-policy: origin-when-cross-origin\\n    server: cloudflare\\n    status: 200\\n    strict-transport-security: max-age=31536000; includeSubDomains; preload\\n    vary: User-Agent\\n    vary: Accept-Encoding\\n    x-b3-sampled: 0\\n    x-content-type-options: nosniff\\n    x-mode: ro\\n    x-node: www_all\\n    x-node: 10-69-179-105-uswest2bprod-9c0a6478-895a-11ea-98c5-b6d34d770\\n    x-proxied: 10-69-159-164-uswest2bprod\\n    x-routing-service: 10-69-187-145-uswest2bprod; site=www\\n    x-xss-protection: 1; report=https://www.yelp.com/xss_protection_report\\n    x-zipkin-id: 9a87fa4730749a04\\n\\\"\\\"\\\"\\nheaders = copyheaders.headers_raw_to_dict(headers_str)\\nrestaurant_list_url = (\\n    lambda index: f\\\"https://www.yelp.com/search/snippet?find_desc=&find_loc=Los%20Angeles%2C%20CA&start={index}\\\"\\n)\\ntotal_number_of_restaurants = 240\\nyelp_raw_data = []\\nfor i in tqdm(range((total_number_of_restaurants // 10) + 1)):\\n    index = (\\n        i * 10 if i * 10 < total_number_of_restaurants else total_number_of_restaurants\\n    )\\n    retries, max_retries = 0, 5e2\\n    while retries < max_retries:\\n        retries += 1\\n        page = requests.get(restaurant_list_url(index), headers=headers)\\n        try:\\n            if page.ok:\\n                yelp_raw_data += json.loads(page.content)[\\\"searchPageProps\\\"][\\n                    \\\"mainContentComponentsListProps\\\"\\n                ]\\n        except:\\n            if retries % 10 == 0:\\n                print(f\\\"Number of attempts to get data for {index}: {retries}\\\")\\n            continue\\n        if retries > max_retries and not page.ok:\\n            print(f\\\"Couldn't get data for index: {index}\\\")\\n        break\";\n",
       "                var nbb_formatted_code = \"headers_str = b\\\"\\\"\\\"\\n    cache-control: max-age=0, must-revalidate, no-cache, no-store, private\\n    cache-control: no-transform\\n    cf-cache-status: DYNAMIC\\n    cf-ray: 58b26184fbd76c86-SJC\\n    cf-request-id: 02635b471c00006c86b019b200000001\\n    content-encoding: gzip\\n    content-security-policy: report-uri https://www.yelp.com/csp_block?id=bf59639897830a99&page=enforced_by_default_directives&policy_hash=7b6f2d6630868fdb2698dac44731677c&site=www&timestamp=1588093661; object-src 'self'; base-uri 'self' https://*.yelpcdn.com https://*.adsrvr.org https://6372968.fls.doubleclick.net; font-src data: 'self' https://*.yelp.com https://*.yelpcdn.com https://fonts.gstatic.com https://connect.facebook.net https://cdnjs.cloudflare.com https://apis.google.com https://www.google-analytics.com https://use.typekit.net https://player.ooyala.com https://use.fontawesome.com https://maxcdn.bootstrapcdn.com https://fonts.googleapis.com\\n    content-security-policy-report-only: report-uri https://www.yelp.com/csp_report_only?id=bf59639897830a99&page=csp_report_frame_directives%2Cfull_site_ssl_csp_report_directives&policy_hash=9dd00a1a6fbb402584b7ce0c1fdb4d14&site=www&timestamp=1588093661; frame-ancestors 'self' https://*.yelp.com; default-src https:; img-src https: data: https://*.adsrvr.org; script-src https: data: 'unsafe-inline' 'unsafe-eval' blob:; style-src https: 'unsafe-inline' data:; connect-src https:; font-src data: 'self' https://*.yelp.com https://*.yelpcdn.com https://fonts.gstatic.com https://connect.facebook.net https://cdnjs.cloudflare.com https://apis.google.com https://www.google-analytics.com https://use.typekit.net https://player.ooyala.com https://use.fontawesome.com https://maxcdn.bootstrapcdn.com https://fonts.googleapis.com; frame-src https: yelp-webview://* yelp://* data:; child-src https: yelp-webview://* yelp://*; media-src https:; object-src 'self'; base-uri 'self' https://*.yelpcdn.com https://*.adsrvr.org https://6372968.fls.doubleclick.net; form-action https: 'self'\\n    content-type: application/json; charset=utf-8\\n    date: Tue, 28 Apr 2020 17:07:42 GMT\\n    expect-ct: max-age=604800, report-uri=\\\"https://report-uri.cloudflare.com/cdn-cgi/beacon/expect-ct\\\"\\n    expires: Tue, 28 Apr 2020 17:07:41 GMT\\n    pragma: no-cache\\n    referrer-policy: origin-when-cross-origin\\n    server: cloudflare\\n    status: 200\\n    strict-transport-security: max-age=31536000; includeSubDomains; preload\\n    vary: User-Agent\\n    vary: Accept-Encoding\\n    x-b3-sampled: 0\\n    x-content-type-options: nosniff\\n    x-mode: ro\\n    x-node: www_all\\n    x-node: 10-69-179-105-uswest2bprod-9c0a6478-895a-11ea-98c5-b6d34d770\\n    x-proxied: 10-69-159-164-uswest2bprod\\n    x-routing-service: 10-69-187-145-uswest2bprod; site=www\\n    x-xss-protection: 1; report=https://www.yelp.com/xss_protection_report\\n    x-zipkin-id: 9a87fa4730749a04\\n\\\"\\\"\\\"\\nheaders = copyheaders.headers_raw_to_dict(headers_str)\\nrestaurant_list_url = (\\n    lambda index: f\\\"https://www.yelp.com/search/snippet?find_desc=&find_loc=Los%20Angeles%2C%20CA&start={index}\\\"\\n)\\ntotal_number_of_restaurants = 240\\nyelp_raw_data = []\\nfor i in tqdm(range((total_number_of_restaurants // 10) + 1)):\\n    index = (\\n        i * 10 if i * 10 < total_number_of_restaurants else total_number_of_restaurants\\n    )\\n    retries, max_retries = 0, 5e2\\n    while retries < max_retries:\\n        retries += 1\\n        page = requests.get(restaurant_list_url(index), headers=headers)\\n        try:\\n            if page.ok:\\n                yelp_raw_data += json.loads(page.content)[\\\"searchPageProps\\\"][\\n                    \\\"mainContentComponentsListProps\\\"\\n                ]\\n        except:\\n            if retries % 10 == 0:\\n                print(f\\\"Number of attempts to get data for {index}: {retries}\\\")\\n            continue\\n        if retries > max_retries and not page.ok:\\n            print(f\\\"Couldn't get data for index: {index}\\\")\\n        break\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "headers_str = b\"\"\"\n",
    "    cache-control: max-age=0, must-revalidate, no-cache, no-store, private\n",
    "    cache-control: no-transform\n",
    "    cf-cache-status: DYNAMIC\n",
    "    cf-ray: 58b26184fbd76c86-SJC\n",
    "    cf-request-id: 02635b471c00006c86b019b200000001\n",
    "    content-encoding: gzip\n",
    "    content-security-policy: report-uri https://www.yelp.com/csp_block?id=bf59639897830a99&page=enforced_by_default_directives&policy_hash=7b6f2d6630868fdb2698dac44731677c&site=www&timestamp=1588093661; object-src 'self'; base-uri 'self' https://*.yelpcdn.com https://*.adsrvr.org https://6372968.fls.doubleclick.net; font-src data: 'self' https://*.yelp.com https://*.yelpcdn.com https://fonts.gstatic.com https://connect.facebook.net https://cdnjs.cloudflare.com https://apis.google.com https://www.google-analytics.com https://use.typekit.net https://player.ooyala.com https://use.fontawesome.com https://maxcdn.bootstrapcdn.com https://fonts.googleapis.com\n",
    "    content-security-policy-report-only: report-uri https://www.yelp.com/csp_report_only?id=bf59639897830a99&page=csp_report_frame_directives%2Cfull_site_ssl_csp_report_directives&policy_hash=9dd00a1a6fbb402584b7ce0c1fdb4d14&site=www&timestamp=1588093661; frame-ancestors 'self' https://*.yelp.com; default-src https:; img-src https: data: https://*.adsrvr.org; script-src https: data: 'unsafe-inline' 'unsafe-eval' blob:; style-src https: 'unsafe-inline' data:; connect-src https:; font-src data: 'self' https://*.yelp.com https://*.yelpcdn.com https://fonts.gstatic.com https://connect.facebook.net https://cdnjs.cloudflare.com https://apis.google.com https://www.google-analytics.com https://use.typekit.net https://player.ooyala.com https://use.fontawesome.com https://maxcdn.bootstrapcdn.com https://fonts.googleapis.com; frame-src https: yelp-webview://* yelp://* data:; child-src https: yelp-webview://* yelp://*; media-src https:; object-src 'self'; base-uri 'self' https://*.yelpcdn.com https://*.adsrvr.org https://6372968.fls.doubleclick.net; form-action https: 'self'\n",
    "    content-type: application/json; charset=utf-8\n",
    "    date: Tue, 28 Apr 2020 17:07:42 GMT\n",
    "    expect-ct: max-age=604800, report-uri=\"https://report-uri.cloudflare.com/cdn-cgi/beacon/expect-ct\"\n",
    "    expires: Tue, 28 Apr 2020 17:07:41 GMT\n",
    "    pragma: no-cache\n",
    "    referrer-policy: origin-when-cross-origin\n",
    "    server: cloudflare\n",
    "    status: 200\n",
    "    strict-transport-security: max-age=31536000; includeSubDomains; preload\n",
    "    vary: User-Agent\n",
    "    vary: Accept-Encoding\n",
    "    x-b3-sampled: 0\n",
    "    x-content-type-options: nosniff\n",
    "    x-mode: ro\n",
    "    x-node: www_all\n",
    "    x-node: 10-69-179-105-uswest2bprod-9c0a6478-895a-11ea-98c5-b6d34d770\n",
    "    x-proxied: 10-69-159-164-uswest2bprod\n",
    "    x-routing-service: 10-69-187-145-uswest2bprod; site=www\n",
    "    x-xss-protection: 1; report=https://www.yelp.com/xss_protection_report\n",
    "    x-zipkin-id: 9a87fa4730749a04\n",
    "\"\"\"\n",
    "headers = copyheaders.headers_raw_to_dict(headers_str)\n",
    "restaurant_list_url = (\n",
    "    lambda index: f\"https://www.yelp.com/search/snippet?find_desc=&find_loc=Los%20Angeles%2C%20CA&start={index}\"\n",
    ")\n",
    "total_number_of_restaurants = 240\n",
    "yelp_raw_data = []\n",
    "for i in tqdm(range((total_number_of_restaurants // 10) + 1)):\n",
    "    index = (\n",
    "        i * 10 if i * 10 < total_number_of_restaurants else total_number_of_restaurants\n",
    "    )\n",
    "    retries, max_retries = 0, 5e2\n",
    "    while retries < max_retries:\n",
    "        retries += 1\n",
    "        page = requests.get(restaurant_list_url(index), headers=headers)\n",
    "        try:\n",
    "            if page.ok:\n",
    "                yelp_raw_data += json.loads(page.content)[\"searchPageProps\"][\n",
    "                    \"mainContentComponentsListProps\"\n",
    "                ]\n",
    "        except:\n",
    "            if retries % 10 == 0:\n",
    "                print(f\"Number of attempts to get data for {index}: {retries}\")\n",
    "            continue\n",
    "        if retries > max_retries and not page.ok:\n",
    "            print(f\"Couldn't get data for index: {index}\")\n",
    "        break"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "id": "fossil-criticism",
   "metadata": {
    "scrolled": false
   },
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "100%|██████████| 480/480 [00:00<00:00, 33043.36it/s]\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>(searchActions,)</th>\n",
       "      <th>(isYelpGuaranteed,)</th>\n",
       "      <th>(bizId,)</th>\n",
       "      <th>(tags,)</th>\n",
       "      <th>(scrollablePhotos, allPhotosHref)</th>\n",
       "      <th>(scrollablePhotos, photoHref)</th>\n",
       "      <th>(scrollablePhotos, photoList)</th>\n",
       "      <th>(scrollablePhotos, isResponsive)</th>\n",
       "      <th>(scrollablePhotos, isScrollable)</th>\n",
       "      <th>(searchResultLayoutType,)</th>\n",
       "      <th>...</th>\n",
       "      <th>(snippet, thumbnail, src)</th>\n",
       "      <th>(snippet, thumbnail, srcset)</th>\n",
       "      <th>(snippet, readMoreText)</th>\n",
       "      <th>(snippet, text)</th>\n",
       "      <th>(searchResultBusiness, parentBusiness, businessUrl)</th>\n",
       "      <th>(searchResultBusiness, parentBusiness, name)</th>\n",
       "      <th>(searchResultBusinessHighlights, bizSiteUrl)</th>\n",
       "      <th>(searchResultBusinessHighlights, businessHighlights)</th>\n",
       "      <th>(childrenBusinessInfo, businessUrls)</th>\n",
       "      <th>(childrenBusinessInfo, businessNames)</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>0</th>\n",
       "      <td>[]</td>\n",
       "      <td>False</td>\n",
       "      <td>rF7KNmSv5sYbwd3D5sA_vw</td>\n",
       "      <td>[{'label': {'color': 'normal', 'text': 'New on...</td>\n",
       "      <td>/biz_photos/rF7KNmSv5sYbwd3D5sA_vw</td>\n",
       "      <td>/adredir?ad_business_id=rF7KNmSv5sYbwd3D5sA_vw...</td>\n",
       "      <td>[{'src': 'https://s3-media0.fl.yelpcdn.com/bph...</td>\n",
       "      <td>True</td>\n",
       "      <td>True</td>\n",
       "      <td>scrollablePhotos</td>\n",
       "      <td>...</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>[]</td>\n",
       "      <td>False</td>\n",
       "      <td>7O1ORGY36A-2aIENyaJWPg</td>\n",
       "      <td>[]</td>\n",
       "      <td>/biz_photos/7O1ORGY36A-2aIENyaJWPg</td>\n",
       "      <td>/biz/howlin-rays-los-angeles-3</td>\n",
       "      <td>[{'src': 'https://s3-media0.fl.yelpcdn.com/bph...</td>\n",
       "      <td>True</td>\n",
       "      <td>True</td>\n",
       "      <td>scrollablePhotos</td>\n",
       "      <td>...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/22VFkvu...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/22VFkvu...</td>\n",
       "      <td>more</td>\n",
       "      <td>I FINALLY got to try this place... but sadly b...</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>[]</td>\n",
       "      <td>False</td>\n",
       "      <td>KQBGm5G8IDkE8LeNY45mbA</td>\n",
       "      <td>[]</td>\n",
       "      <td>/biz_photos/KQBGm5G8IDkE8LeNY45mbA</td>\n",
       "      <td>/biz/wurstk%C3%BCche-los-angeles-2</td>\n",
       "      <td>[{'src': 'https://s3-media0.fl.yelpcdn.com/bph...</td>\n",
       "      <td>True</td>\n",
       "      <td>True</td>\n",
       "      <td>scrollablePhotos</td>\n",
       "      <td>...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/mUmrPnL...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/mUmrPnL...</td>\n",
       "      <td>more</td>\n",
       "      <td>A group of us stopped by for dinner before exp...</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>[{'content': {'text': {'text': 'Offers takeout...</td>\n",
       "      <td>False</td>\n",
       "      <td>iSZpZgVnASwEmlq0DORY2A</td>\n",
       "      <td>[]</td>\n",
       "      <td>/biz_photos/iSZpZgVnASwEmlq0DORY2A</td>\n",
       "      <td>/biz/daikokuya-little-tokyo-los-angeles</td>\n",
       "      <td>[{'src': 'https://s3-media0.fl.yelpcdn.com/bph...</td>\n",
       "      <td>True</td>\n",
       "      <td>True</td>\n",
       "      <td>scrollablePhotos</td>\n",
       "      <td>...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/5P7QP1p...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/5P7QP1p...</td>\n",
       "      <td>more</td>\n",
       "      <td>Daikokuya just never disappoints!\\nWent last n...</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>[{'content': {'text': {'text': 'Offers takeout...</td>\n",
       "      <td>False</td>\n",
       "      <td>MlmcOkwaNnxl3Zuk6HsPCQ</td>\n",
       "      <td>[{'label': {'color': 'normal', 'text': 'Curren...</td>\n",
       "      <td>/biz_photos/MlmcOkwaNnxl3Zuk6HsPCQ</td>\n",
       "      <td>/biz/slurpin-ramen-bar-los-angeles-los-angeles</td>\n",
       "      <td>[{'src': 'https://s3-media0.fl.yelpcdn.com/bph...</td>\n",
       "      <td>True</td>\n",
       "      <td>True</td>\n",
       "      <td>scrollablePhotos</td>\n",
       "      <td>...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/-6T8kS9...</td>\n",
       "      <td>https://s3-media0.fl.yelpcdn.com/photo/-6T8kS9...</td>\n",
       "      <td>more</td>\n",
       "      <td>Covid delivery review:\\nOooooooof this hit the...</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "      <td>NaN</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "<p>5 rows × 53 columns</p>\n",
       "</div>"
      ],
      "text/plain": [
       "                                    (searchActions,)  (isYelpGuaranteed,)  \\\n",
       "0                                                 []                False   \n",
       "1                                                 []                False   \n",
       "2                                                 []                False   \n",
       "3  [{'content': {'text': {'text': 'Offers takeout...                False   \n",
       "4  [{'content': {'text': {'text': 'Offers takeout...                False   \n",
       "\n",
       "                 (bizId,)                                            (tags,)  \\\n",
       "0  rF7KNmSv5sYbwd3D5sA_vw  [{'label': {'color': 'normal', 'text': 'New on...   \n",
       "1  7O1ORGY36A-2aIENyaJWPg                                                 []   \n",
       "2  KQBGm5G8IDkE8LeNY45mbA                                                 []   \n",
       "3  iSZpZgVnASwEmlq0DORY2A                                                 []   \n",
       "4  MlmcOkwaNnxl3Zuk6HsPCQ  [{'label': {'color': 'normal', 'text': 'Curren...   \n",
       "\n",
       "    (scrollablePhotos, allPhotosHref)  \\\n",
       "0  /biz_photos/rF7KNmSv5sYbwd3D5sA_vw   \n",
       "1  /biz_photos/7O1ORGY36A-2aIENyaJWPg   \n",
       "2  /biz_photos/KQBGm5G8IDkE8LeNY45mbA   \n",
       "3  /biz_photos/iSZpZgVnASwEmlq0DORY2A   \n",
       "4  /biz_photos/MlmcOkwaNnxl3Zuk6HsPCQ   \n",
       "\n",
       "                       (scrollablePhotos, photoHref)  \\\n",
       "0  /adredir?ad_business_id=rF7KNmSv5sYbwd3D5sA_vw...   \n",
       "1                     /biz/howlin-rays-los-angeles-3   \n",
       "2                 /biz/wurstk%C3%BCche-los-angeles-2   \n",
       "3            /biz/daikokuya-little-tokyo-los-angeles   \n",
       "4     /biz/slurpin-ramen-bar-los-angeles-los-angeles   \n",
       "\n",
       "                       (scrollablePhotos, photoList)  \\\n",
       "0  [{'src': 'https://s3-media0.fl.yelpcdn.com/bph...   \n",
       "1  [{'src': 'https://s3-media0.fl.yelpcdn.com/bph...   \n",
       "2  [{'src': 'https://s3-media0.fl.yelpcdn.com/bph...   \n",
       "3  [{'src': 'https://s3-media0.fl.yelpcdn.com/bph...   \n",
       "4  [{'src': 'https://s3-media0.fl.yelpcdn.com/bph...   \n",
       "\n",
       "   (scrollablePhotos, isResponsive)  (scrollablePhotos, isScrollable)  \\\n",
       "0                              True                              True   \n",
       "1                              True                              True   \n",
       "2                              True                              True   \n",
       "3                              True                              True   \n",
       "4                              True                              True   \n",
       "\n",
       "  (searchResultLayoutType,)  ...  \\\n",
       "0          scrollablePhotos  ...   \n",
       "1          scrollablePhotos  ...   \n",
       "2          scrollablePhotos  ...   \n",
       "3          scrollablePhotos  ...   \n",
       "4          scrollablePhotos  ...   \n",
       "\n",
       "                           (snippet, thumbnail, src)  \\\n",
       "0                                                NaN   \n",
       "1  https://s3-media0.fl.yelpcdn.com/photo/22VFkvu...   \n",
       "2  https://s3-media0.fl.yelpcdn.com/photo/mUmrPnL...   \n",
       "3  https://s3-media0.fl.yelpcdn.com/photo/5P7QP1p...   \n",
       "4  https://s3-media0.fl.yelpcdn.com/photo/-6T8kS9...   \n",
       "\n",
       "                        (snippet, thumbnail, srcset) (snippet, readMoreText)  \\\n",
       "0                                                NaN                     NaN   \n",
       "1  https://s3-media0.fl.yelpcdn.com/photo/22VFkvu...                    more   \n",
       "2  https://s3-media0.fl.yelpcdn.com/photo/mUmrPnL...                    more   \n",
       "3  https://s3-media0.fl.yelpcdn.com/photo/5P7QP1p...                    more   \n",
       "4  https://s3-media0.fl.yelpcdn.com/photo/-6T8kS9...                    more   \n",
       "\n",
       "                                     (snippet, text)  \\\n",
       "0                                                NaN   \n",
       "1  I FINALLY got to try this place... but sadly b...   \n",
       "2  A group of us stopped by for dinner before exp...   \n",
       "3  Daikokuya just never disappoints!\\nWent last n...   \n",
       "4  Covid delivery review:\\nOooooooof this hit the...   \n",
       "\n",
       "  (searchResultBusiness, parentBusiness, businessUrl)  \\\n",
       "0                                                NaN    \n",
       "1                                                NaN    \n",
       "2                                                NaN    \n",
       "3                                                NaN    \n",
       "4                                                NaN    \n",
       "\n",
       "   (searchResultBusiness, parentBusiness, name)  \\\n",
       "0                                           NaN   \n",
       "1                                           NaN   \n",
       "2                                           NaN   \n",
       "3                                           NaN   \n",
       "4                                           NaN   \n",
       "\n",
       "   (searchResultBusinessHighlights, bizSiteUrl)  \\\n",
       "0                                           NaN   \n",
       "1                                           NaN   \n",
       "2                                           NaN   \n",
       "3                                           NaN   \n",
       "4                                           NaN   \n",
       "\n",
       "  (searchResultBusinessHighlights, businessHighlights)  \\\n",
       "0                                                NaN     \n",
       "1                                                NaN     \n",
       "2                                                NaN     \n",
       "3                                                NaN     \n",
       "4                                                NaN     \n",
       "\n",
       "  (childrenBusinessInfo, businessUrls) (childrenBusinessInfo, businessNames)  \n",
       "0                                  NaN                                   NaN  \n",
       "1                                  NaN                                   NaN  \n",
       "2                                  NaN                                   NaN  \n",
       "3                                  NaN                                   NaN  \n",
       "4                                  NaN                                   NaN  \n",
       "\n",
       "[5 rows x 53 columns]"
      ]
     },
     "execution_count": 3,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 91.1 ms (started: 2021-03-09 15:57:21 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 3;\n",
       "                var nbb_unformatted_code = \"yelp_raw = pd.DataFrame(\\n    [flatten(content) for content in tqdm(yelp_raw_data) if \\\"bizId\\\" in content.keys()]\\n)\\nyelp_raw.to_csv(\\\"./data/yelp_raw.csv\\\")\\nyelp_raw.head()\";\n",
       "                var nbb_formatted_code = \"yelp_raw = pd.DataFrame(\\n    [flatten(content) for content in tqdm(yelp_raw_data) if \\\"bizId\\\" in content.keys()]\\n)\\nyelp_raw.to_csv(\\\"./data/yelp_raw.csv\\\")\\nyelp_raw.head()\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_raw = pd.DataFrame(\n",
    "    [flatten(content) for content in tqdm(yelp_raw_data) if \"bizId\" in content.keys()]\n",
    ")\n",
    "yelp_raw.to_csv(\"./data/yelp_raw.csv\")\n",
    "yelp_raw.head()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "knowing-island",
   "metadata": {},
   "source": [
    "---\n",
    "## Data Pre-processing\n",
    "\n",
    "Let's take out some of the unecessary columns for now."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "id": "alert-xerox",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "Index([                                       ('searchActions',),\n",
       "                                           ('isYelpGuaranteed',),\n",
       "                                                      ('bizId',),\n",
       "                                                       ('tags',),\n",
       "                           ('scrollablePhotos', 'allPhotosHref'),\n",
       "                               ('scrollablePhotos', 'photoHref'),\n",
       "                               ('scrollablePhotos', 'photoList'),\n",
       "                            ('scrollablePhotos', 'isResponsive'),\n",
       "                            ('scrollablePhotos', 'isScrollable'),\n",
       "                                     ('searchResultLayoutType',),\n",
       "                             ('searchResultBusinessHighlights',),\n",
       "                                      ('verifiedLicenseLayout',),\n",
       "                                           ('serviceOfferings',),\n",
       "                                                    ('snippet',),\n",
       "                                                  ('markerKey',),\n",
       "                                       ('adLoggingInfo', 'slot'),\n",
       "                              ('adLoggingInfo', 'placementSlot'),\n",
       "                                       ('adLoggingInfo', 'flow'),\n",
       "                              ('adLoggingInfo', 'opportunityId'),\n",
       "                               ('adLoggingInfo', 'adCampaignId'),\n",
       "                               ('adLoggingInfo', 'isShowcaseAd'),\n",
       "                                  ('adLoggingInfo', 'placement'),\n",
       "                                       ('offerCampaignDetails',),\n",
       "                      ('searchResultBusinessPortfolioProjects',),\n",
       "                      ('searchResultBusiness', 'parentBusiness'),\n",
       "                             ('searchResultBusiness', 'ranking'),\n",
       "                         ('searchResultBusiness', 'reviewCount'),\n",
       "                        ('searchResultBusiness', 'renderAdInfo'),\n",
       "                                ('searchResultBusiness', 'name'),\n",
       "                       ('searchResultBusiness', 'neighborhoods'),\n",
       "                              ('searchResultBusiness', 'rating'),\n",
       "                         ('searchResultBusiness', 'businessUrl'),\n",
       "                                ('searchResultBusiness', 'isAd'),\n",
       "                         ('searchResultBusiness', 'serviceArea'),\n",
       "                               ('searchResultBusiness', 'phone'),\n",
       "                          ('searchResultBusiness', 'priceRange'),\n",
       "                      ('searchResultBusiness', 'alternateNames'),\n",
       "                    ('searchResultBusiness', 'formattedAddress'),\n",
       "                      ('searchResultBusiness', 'servicePricing'),\n",
       "                          ('searchResultBusiness', 'bizSiteUrl'),\n",
       "                          ('searchResultBusiness', 'categories'),\n",
       "                                       ('childrenBusinessInfo',),\n",
       "                                      ('snippet', 'readMoreUrl'),\n",
       "                                 ('snippet', 'thumbnail', 'src'),\n",
       "                              ('snippet', 'thumbnail', 'srcset'),\n",
       "                                     ('snippet', 'readMoreText'),\n",
       "                                             ('snippet', 'text'),\n",
       "       ('searchResultBusiness', 'parentBusiness', 'businessUrl'),\n",
       "              ('searchResultBusiness', 'parentBusiness', 'name'),\n",
       "                ('searchResultBusinessHighlights', 'bizSiteUrl'),\n",
       "        ('searchResultBusinessHighlights', 'businessHighlights'),\n",
       "                        ('childrenBusinessInfo', 'businessUrls'),\n",
       "                       ('childrenBusinessInfo', 'businessNames')],\n",
       "      dtype='object')"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 7.82 ms (started: 2021-03-09 15:57:21 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 4;\n",
       "                var nbb_unformatted_code = \"yelp_raw.columns\";\n",
       "                var nbb_formatted_code = \"yelp_raw.columns\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_raw.columns"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "id": "numerous-hacker",
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "/usr/local/lib/python3.8/site-packages/numpy/core/_asarray.py:102: VisibleDeprecationWarning: Creating an ndarray from ragged nested sequences (which is a list-or-tuple of lists-or-tuples-or ndarrays with different lengths or shapes) is deprecated. If you meant to do this, you must specify 'dtype=object' when creating the ndarray.\n",
      "  return array(a, dtype, copy=False, order=order)\n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>reviewCount</th>\n",
       "      <th>renderAdInfo</th>\n",
       "      <th>name</th>\n",
       "      <th>neighborhoods</th>\n",
       "      <th>rating</th>\n",
       "      <th>businessUrl</th>\n",
       "      <th>isAd</th>\n",
       "      <th>phone</th>\n",
       "      <th>priceRange</th>\n",
       "      <th>alternateNames</th>\n",
       "      <th>formattedAddress</th>\n",
       "      <th>categories</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>6636</td>\n",
       "      <td>False</td>\n",
       "      <td>Howlin’ Ray’s</td>\n",
       "      <td>[Chinatown]</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/howlin-rays-los-angeles-3</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 935-8399</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>727 N Broadway</td>\n",
       "      <td>[{'url': '/search?cflt=southern&amp;find_loc=Los+A...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>8449</td>\n",
       "      <td>False</td>\n",
       "      <td>Wurstküche</td>\n",
       "      <td>[Arts District]</td>\n",
       "      <td>4.0</td>\n",
       "      <td>/biz/wurstk%C3%BCche-los-angeles-2</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 687-4444</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>800 E 3rd St</td>\n",
       "      <td>[{'url': '/search?cflt=hotdog&amp;find_loc=Los+Ang...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>8644</td>\n",
       "      <td>False</td>\n",
       "      <td>Daikokuya Little Tokyo</td>\n",
       "      <td>[Little Tokyo]</td>\n",
       "      <td>4.0</td>\n",
       "      <td>/biz/daikokuya-little-tokyo-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 626-1680</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>327 E 1st St</td>\n",
       "      <td>[{'url': '/search?cflt=ramen&amp;find_loc=Los+Ange...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>4641</td>\n",
       "      <td>False</td>\n",
       "      <td>Slurpin’ Ramen Bar - Los Angeles</td>\n",
       "      <td>[Koreatown]</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/slurpin-ramen-bar-los-angeles-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 388-8607</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>3500 W 8th St</td>\n",
       "      <td>[{'url': '/search?cflt=ramen&amp;find_loc=Los+Ange...</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>5</th>\n",
       "      <td>6228</td>\n",
       "      <td>False</td>\n",
       "      <td>Bestia</td>\n",
       "      <td>[Downtown]</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/bestia-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 514-5724</td>\n",
       "      <td>$$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>2121 E 7th Pl</td>\n",
       "      <td>[{'url': '/search?cflt=italian&amp;find_loc=Los+An...</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "   reviewCount  renderAdInfo                              name  \\\n",
       "1         6636         False                     Howlin’ Ray’s   \n",
       "2         8449         False                        Wurstküche   \n",
       "3         8644         False            Daikokuya Little Tokyo   \n",
       "4         4641         False  Slurpin’ Ramen Bar - Los Angeles   \n",
       "5         6228         False                            Bestia   \n",
       "\n",
       "     neighborhoods  rating                                     businessUrl  \\\n",
       "1      [Chinatown]     4.5                  /biz/howlin-rays-los-angeles-3   \n",
       "2  [Arts District]     4.0              /biz/wurstk%C3%BCche-los-angeles-2   \n",
       "3   [Little Tokyo]     4.0         /biz/daikokuya-little-tokyo-los-angeles   \n",
       "4      [Koreatown]     4.5  /biz/slurpin-ramen-bar-los-angeles-los-angeles   \n",
       "5       [Downtown]     4.5                         /biz/bestia-los-angeles   \n",
       "\n",
       "    isAd           phone priceRange alternateNames formattedAddress  \\\n",
       "1  False  (213) 935-8399         $$             []   727 N Broadway   \n",
       "2  False  (213) 687-4444         $$             []     800 E 3rd St   \n",
       "3  False  (213) 626-1680         $$             []     327 E 1st St   \n",
       "4  False  (213) 388-8607         $$             []    3500 W 8th St   \n",
       "5  False  (213) 514-5724        $$$             []    2121 E 7th Pl   \n",
       "\n",
       "                                          categories  \n",
       "1  [{'url': '/search?cflt=southern&find_loc=Los+A...  \n",
       "2  [{'url': '/search?cflt=hotdog&find_loc=Los+Ang...  \n",
       "3  [{'url': '/search?cflt=ramen&find_loc=Los+Ange...  \n",
       "4  [{'url': '/search?cflt=ramen&find_loc=Los+Ange...  \n",
       "5  [{'url': '/search?cflt=italian&find_loc=Los+An...  "
      ]
     },
     "execution_count": 5,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 29.9 ms (started: 2021-03-09 15:57:21 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 5;\n",
       "                var nbb_unformatted_code = \"yelp_filtered = yelp_raw[\\n    [col for col in yelp_raw.columns if \\\"searchResultBusiness\\\" in col]\\n].dropna(axis=1)\\nyelp_filtered = yelp_filtered[\\n    yelp_filtered[(\\\"searchResultBusiness\\\", \\\"priceRange\\\")] != \\\"\\\"\\n]  # Remove rows with no price range\\nyelp_filtered.columns = [tuple(col)[-1] for col in yelp_filtered.columns]\\nyelp_filtered.to_csv(\\\"./data/yelp_filtered.csv\\\")\\nyelp_filtered.head()\";\n",
       "                var nbb_formatted_code = \"yelp_filtered = yelp_raw[\\n    [col for col in yelp_raw.columns if \\\"searchResultBusiness\\\" in col]\\n].dropna(axis=1)\\nyelp_filtered = yelp_filtered[\\n    yelp_filtered[(\\\"searchResultBusiness\\\", \\\"priceRange\\\")] != \\\"\\\"\\n]  # Remove rows with no price range\\nyelp_filtered.columns = [tuple(col)[-1] for col in yelp_filtered.columns]\\nyelp_filtered.to_csv(\\\"./data/yelp_filtered.csv\\\")\\nyelp_filtered.head()\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_filtered = yelp_raw[\n",
    "    [col for col in yelp_raw.columns if \"searchResultBusiness\" in col]\n",
    "].dropna(axis=1)\n",
    "yelp_filtered = yelp_filtered[\n",
    "    yelp_filtered[(\"searchResultBusiness\", \"priceRange\")] != \"\"\n",
    "]  # Remove rows with no price range\n",
    "yelp_filtered.columns = [tuple(col)[-1] for col in yelp_filtered.columns]\n",
    "yelp_filtered.to_csv(\"./data/yelp_filtered.csv\")\n",
    "yelp_filtered.head()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "artistic-lender",
   "metadata": {},
   "source": [
    "Add the lat, long for each address"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "id": "historical-confidence",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>reviewCount</th>\n",
       "      <th>renderAdInfo</th>\n",
       "      <th>name</th>\n",
       "      <th>neighborhoods</th>\n",
       "      <th>rating</th>\n",
       "      <th>businessUrl</th>\n",
       "      <th>isAd</th>\n",
       "      <th>phone</th>\n",
       "      <th>priceRange</th>\n",
       "      <th>alternateNames</th>\n",
       "      <th>formattedAddress</th>\n",
       "      <th>categories</th>\n",
       "      <th>latitude</th>\n",
       "      <th>longitude</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>6636</td>\n",
       "      <td>False</td>\n",
       "      <td>Howlin’ Ray’s</td>\n",
       "      <td>[Chinatown]</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/howlin-rays-los-angeles-3</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 935-8399</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>727 N Broadway</td>\n",
       "      <td>[{'url': '/search?cflt=southern&amp;find_loc=Los+A...</td>\n",
       "      <td>34.061519</td>\n",
       "      <td>-118.239473</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>8449</td>\n",
       "      <td>False</td>\n",
       "      <td>Wurstküche</td>\n",
       "      <td>[Arts District]</td>\n",
       "      <td>4.0</td>\n",
       "      <td>/biz/wurstk%C3%BCche-los-angeles-2</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 687-4444</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>800 E 3rd St</td>\n",
       "      <td>[{'url': '/search?cflt=hotdog&amp;find_loc=Los+Ang...</td>\n",
       "      <td>33.896347</td>\n",
       "      <td>-118.117083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>8644</td>\n",
       "      <td>False</td>\n",
       "      <td>Daikokuya Little Tokyo</td>\n",
       "      <td>[Little Tokyo]</td>\n",
       "      <td>4.0</td>\n",
       "      <td>/biz/daikokuya-little-tokyo-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 626-1680</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>327 E 1st St</td>\n",
       "      <td>[{'url': '/search?cflt=ramen&amp;find_loc=Los+Ange...</td>\n",
       "      <td>34.049971</td>\n",
       "      <td>-118.240083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>4641</td>\n",
       "      <td>False</td>\n",
       "      <td>Slurpin’ Ramen Bar - Los Angeles</td>\n",
       "      <td>[Koreatown]</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/slurpin-ramen-bar-los-angeles-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 388-8607</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>3500 W 8th St</td>\n",
       "      <td>[{'url': '/search?cflt=ramen&amp;find_loc=Los+Ange...</td>\n",
       "      <td>34.057590</td>\n",
       "      <td>-118.306725</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>5</th>\n",
       "      <td>6228</td>\n",
       "      <td>False</td>\n",
       "      <td>Bestia</td>\n",
       "      <td>[Downtown]</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/bestia-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 514-5724</td>\n",
       "      <td>$$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>2121 E 7th Pl</td>\n",
       "      <td>[{'url': '/search?cflt=italian&amp;find_loc=Los+An...</td>\n",
       "      <td>34.033738</td>\n",
       "      <td>-118.229309</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "   reviewCount  renderAdInfo                              name  \\\n",
       "1         6636         False                     Howlin’ Ray’s   \n",
       "2         8449         False                        Wurstküche   \n",
       "3         8644         False            Daikokuya Little Tokyo   \n",
       "4         4641         False  Slurpin’ Ramen Bar - Los Angeles   \n",
       "5         6228         False                            Bestia   \n",
       "\n",
       "     neighborhoods  rating                                     businessUrl  \\\n",
       "1      [Chinatown]     4.5                  /biz/howlin-rays-los-angeles-3   \n",
       "2  [Arts District]     4.0              /biz/wurstk%C3%BCche-los-angeles-2   \n",
       "3   [Little Tokyo]     4.0         /biz/daikokuya-little-tokyo-los-angeles   \n",
       "4      [Koreatown]     4.5  /biz/slurpin-ramen-bar-los-angeles-los-angeles   \n",
       "5       [Downtown]     4.5                         /biz/bestia-los-angeles   \n",
       "\n",
       "    isAd           phone priceRange alternateNames formattedAddress  \\\n",
       "1  False  (213) 935-8399         $$             []   727 N Broadway   \n",
       "2  False  (213) 687-4444         $$             []     800 E 3rd St   \n",
       "3  False  (213) 626-1680         $$             []     327 E 1st St   \n",
       "4  False  (213) 388-8607         $$             []    3500 W 8th St   \n",
       "5  False  (213) 514-5724        $$$             []    2121 E 7th Pl   \n",
       "\n",
       "                                          categories   latitude   longitude  \n",
       "1  [{'url': '/search?cflt=southern&find_loc=Los+A...  34.061519 -118.239473  \n",
       "2  [{'url': '/search?cflt=hotdog&find_loc=Los+Ang...  33.896347 -118.117083  \n",
       "3  [{'url': '/search?cflt=ramen&find_loc=Los+Ange...  34.049971 -118.240083  \n",
       "4  [{'url': '/search?cflt=ramen&find_loc=Los+Ange...  34.057590 -118.306725  \n",
       "5  [{'url': '/search?cflt=italian&find_loc=Los+An...  34.033738 -118.229309  "
      ]
     },
     "execution_count": 6,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 2min 34s (started: 2021-03-09 15:57:21 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 6;\n",
       "                var nbb_unformatted_code = \"locator = Nominatim(user_agent=\\\"myGeocoder\\\")\\n\\n\\ndef get_latlong(address: str, locator=locator):\\n    \\\"\\\"\\\"Get lat, long from string address\\\"\\\"\\\"\\n    location = locator.geocode(f\\\"{address}, Los Angeles, CA\\\")\\n    try:\\n        lat, long = location.latitude, location.longitude\\n    except:\\n        lat, long = None, None\\n    return lat, long\\n\\n\\nyelp_filtered = pd.concat(\\n    [\\n        yelp_filtered,\\n        pd.DataFrame(\\n            yelp_filtered[\\\"formattedAddress\\\"]\\n            .apply(lambda address: get_latlong(address))\\n            .tolist(),\\n            columns=[\\\"latitude\\\", \\\"longitude\\\"],\\n            index=yelp_filtered.index,\\n        ),\\n    ],\\n    axis=1,\\n)\\nyelp_filtered.head()\";\n",
       "                var nbb_formatted_code = \"locator = Nominatim(user_agent=\\\"myGeocoder\\\")\\n\\n\\ndef get_latlong(address: str, locator=locator):\\n    \\\"\\\"\\\"Get lat, long from string address\\\"\\\"\\\"\\n    location = locator.geocode(f\\\"{address}, Los Angeles, CA\\\")\\n    try:\\n        lat, long = location.latitude, location.longitude\\n    except:\\n        lat, long = None, None\\n    return lat, long\\n\\n\\nyelp_filtered = pd.concat(\\n    [\\n        yelp_filtered,\\n        pd.DataFrame(\\n            yelp_filtered[\\\"formattedAddress\\\"]\\n            .apply(lambda address: get_latlong(address))\\n            .tolist(),\\n            columns=[\\\"latitude\\\", \\\"longitude\\\"],\\n            index=yelp_filtered.index,\\n        ),\\n    ],\\n    axis=1,\\n)\\nyelp_filtered.head()\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "locator = Nominatim(user_agent=\"myGeocoder\")\n",
    "\n",
    "\n",
    "def get_latlong(address: str, locator=locator):\n",
    "    \"\"\"Get lat, long from string address\"\"\"\n",
    "    location = locator.geocode(f\"{address}, Los Angeles, CA\")\n",
    "    try:\n",
    "        lat, long = location.latitude, location.longitude\n",
    "    except:\n",
    "        lat, long = None, None\n",
    "    return lat, long\n",
    "\n",
    "\n",
    "yelp_filtered = pd.concat(\n",
    "    [\n",
    "        yelp_filtered,\n",
    "        pd.DataFrame(\n",
    "            yelp_filtered[\"formattedAddress\"]\n",
    "            .apply(lambda address: get_latlong(address))\n",
    "            .tolist(),\n",
    "            columns=[\"latitude\", \"longitude\"],\n",
    "            index=yelp_filtered.index,\n",
    "        ),\n",
    "    ],\n",
    "    axis=1,\n",
    ")\n",
    "yelp_filtered.head()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aware-inventory",
   "metadata": {},
   "source": [
    "Convert neighbourhoods into a single value for each row"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "id": "animated-ceiling",
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "data": {
      "text/plain": [
       "1    234\n",
       "Name: neighborhoods, dtype: int64"
      ]
     },
     "execution_count": 7,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 4.93 ms (started: 2021-03-09 15:59:55 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 7;\n",
       "                var nbb_unformatted_code = \"yelp_filtered[\\\"neighborhoods\\\"].apply(\\n    lambda neighbourhoods: len(neighbourhoods)\\n).value_counts()\";\n",
       "                var nbb_formatted_code = \"yelp_filtered[\\\"neighborhoods\\\"].apply(\\n    lambda neighbourhoods: len(neighbourhoods)\\n).value_counts()\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_filtered[\"neighborhoods\"].apply(\n",
    "    lambda neighbourhoods: len(neighbourhoods)\n",
    ").value_counts()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "id": "stuck-imaging",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 1.31 ms (started: 2021-03-09 15:59:55 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 8;\n",
       "                var nbb_unformatted_code = \"yelp_filtered[\\\"neighborhoods\\\"] = yelp_filtered[\\\"neighborhoods\\\"].apply(\\n    lambda neighbourhoods: neighbourhoods[0]\\n)\";\n",
       "                var nbb_formatted_code = \"yelp_filtered[\\\"neighborhoods\\\"] = yelp_filtered[\\\"neighborhoods\\\"].apply(\\n    lambda neighbourhoods: neighbourhoods[0]\\n)\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_filtered[\"neighborhoods\"] = yelp_filtered[\"neighborhoods\"].apply(\n",
    "    lambda neighbourhoods: neighbourhoods[0]\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "id": "adult-fabric",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "Downtown                   30\n",
       "Koreatown                  23\n",
       "Hollywood                  15\n",
       "Little Tokyo               15\n",
       "Wilshire Center            14\n",
       "Beverly Grove              13\n",
       "Fairfax                    11\n",
       "Mid-Wilshire               10\n",
       "East Hollywood             10\n",
       "Los Feliz                   9\n",
       "Silver Lake                 6\n",
       "Arts District               6\n",
       "Chinatown                   5\n",
       "Hancock Park                5\n",
       "Echo Park                   4\n",
       "Sawtelle                    4\n",
       "University Park             3\n",
       "Mid-City                    3\n",
       "Larchmont                   3\n",
       "Arlington Heights           3\n",
       "Harvard Heights             3\n",
       "Lincoln Heights             3\n",
       "Palms                       3\n",
       "Highland Park               2\n",
       "Century City                2\n",
       "Mar Vista                   2\n",
       "Pico-Robertson              2\n",
       "Windsor Square              2\n",
       "Atwater Village             2\n",
       "Westlake                    2\n",
       "Boyle Heights               2\n",
       "Exposition Park             2\n",
       "Historic South Central      2\n",
       "Pico-Union                  2\n",
       "Eagle Rock                  2\n",
       "Elysian Park                1\n",
       "Beverlywood                 1\n",
       "Westchester                 1\n",
       "View Park/Windsor Hills     1\n",
       "Studio City                 1\n",
       "West Adams                  1\n",
       "Hollywood Hills West        1\n",
       "Westwood                    1\n",
       "Griffith Park               1\n",
       "Name: neighborhoods, dtype: int64"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 4.17 ms (started: 2021-03-09 15:59:55 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 9;\n",
       "                var nbb_unformatted_code = \"yelp_filtered[\\\"neighborhoods\\\"].value_counts()\";\n",
       "                var nbb_formatted_code = \"yelp_filtered[\\\"neighborhoods\\\"].value_counts()\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_filtered[\"neighborhoods\"].value_counts()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "printable-richmond",
   "metadata": {},
   "source": [
    "What categories do our restaurants have?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "id": "premium-flower",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "Breakfast & Brunch           33\n",
       "Korean                       27\n",
       "Seafood                      21\n",
       "Sandwiches                   21\n",
       "Coffee & Tea                 21\n",
       "American (New)               21\n",
       "Mexican                      20\n",
       "Barbeque                     17\n",
       "Desserts                     15\n",
       "Italian                      14\n",
       "Noodles                      14\n",
       "Japanese                     13\n",
       "Ice Cream & Frozen Yogurt    13\n",
       "Bakeries                     13\n",
       "Cocktail Bars                12\n",
       "Sushi Bars                   11\n",
       "Ramen                        10\n",
       "Pizza                         9\n",
       "American (Traditional)        9\n",
       "Vegan                         9\n",
       "Mediterranean                 8\n",
       "Cafes                         8\n",
       "Salad                         8\n",
       "Burgers                       7\n",
       "Soup                          7\n",
       "Chicken Shop                  7\n",
       "Halal                         6\n",
       "Chinese                       6\n",
       "Comfort Food                  6\n",
       "Thai                          6\n",
       "dtype: int64"
      ]
     },
     "execution_count": 10,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 7.52 ms (started: 2021-03-09 15:59:55 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 10;\n",
       "                var nbb_unformatted_code = \"from collections import Counter\\n\\npd.Series(\\n    [\\n        cat\\n        for cats in yelp_filtered[\\\"categories\\\"]\\n        .apply(\\n            lambda raw_categories: np.array(\\n                [category[\\\"title\\\"] for category in raw_categories]\\n            )\\n        )\\n        .to_numpy()\\n        for cat in cats\\n    ]\\n).value_counts()[:30]\";\n",
       "                var nbb_formatted_code = \"from collections import Counter\\n\\npd.Series(\\n    [\\n        cat\\n        for cats in yelp_filtered[\\\"categories\\\"]\\n        .apply(\\n            lambda raw_categories: np.array(\\n                [category[\\\"title\\\"] for category in raw_categories]\\n            )\\n        )\\n        .to_numpy()\\n        for cat in cats\\n    ]\\n).value_counts()[:30]\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "from collections import Counter\n",
    "\n",
    "pd.Series(\n",
    "    [\n",
    "        cat\n",
    "        for cats in yelp_filtered[\"categories\"]\n",
    "        .apply(\n",
    "            lambda raw_categories: np.array(\n",
    "                [category[\"title\"] for category in raw_categories]\n",
    "            )\n",
    "        )\n",
    "        .to_numpy()\n",
    "        for cat in cats\n",
    "    ]\n",
    ").value_counts()[:30]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "id": "exciting-politics",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 1.82 ms (started: 2021-03-09 15:59:55 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 11;\n",
       "                var nbb_unformatted_code = \"yelp_filtered[\\\"categories\\\"] = yelp_filtered[\\\"categories\\\"].apply(\\n    lambda raw_categories: np.array([category[\\\"title\\\"] for category in raw_categories])\\n)\";\n",
       "                var nbb_formatted_code = \"yelp_filtered[\\\"categories\\\"] = yelp_filtered[\\\"categories\\\"].apply(\\n    lambda raw_categories: np.array([category[\\\"title\\\"] for category in raw_categories])\\n)\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "yelp_filtered[\"categories\"] = yelp_filtered[\"categories\"].apply(\n",
    "    lambda raw_categories: np.array([category[\"title\"] for category in raw_categories])\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 30,
   "id": "facial-economics",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 11.5 ms (started: 2021-03-10 08:00:35 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 30;\n",
       "                var nbb_unformatted_code = \"# Save this dataset\\nyelp_filtered.to_csv(\\\"./data/yelp_cleaned.csv\\\")\";\n",
       "                var nbb_formatted_code = \"# Save this dataset\\nyelp_filtered.to_csv(\\\"./data/yelp_cleaned.csv\\\")\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "# Save this dataset\n",
    "yelp_filtered.to_csv(\"./data/yelp_cleaned.csv\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "automotive-train",
   "metadata": {},
   "source": [
    "---\n",
    "<a id=2></a>\n",
    "## Optimization\n",
    "\n",
    "This can be run indepedently from `Web Scraping Yelp!` as long as `yelp_cleaned.csv` is loaded"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 32,
   "id": "developing-touch",
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>reviewCount</th>\n",
       "      <th>renderAdInfo</th>\n",
       "      <th>name</th>\n",
       "      <th>neighborhoods</th>\n",
       "      <th>rating</th>\n",
       "      <th>businessUrl</th>\n",
       "      <th>isAd</th>\n",
       "      <th>phone</th>\n",
       "      <th>priceRange</th>\n",
       "      <th>alternateNames</th>\n",
       "      <th>formattedAddress</th>\n",
       "      <th>categories</th>\n",
       "      <th>latitude</th>\n",
       "      <th>longitude</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>1</th>\n",
       "      <td>6636</td>\n",
       "      <td>False</td>\n",
       "      <td>Howlin’ Ray’s</td>\n",
       "      <td>Chinatown</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/howlin-rays-los-angeles-3</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 935-8399</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>727 N Broadway</td>\n",
       "      <td>['Southern' 'Chicken Shop' 'American (Traditio...</td>\n",
       "      <td>34.061519</td>\n",
       "      <td>-118.239473</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>2</th>\n",
       "      <td>8449</td>\n",
       "      <td>False</td>\n",
       "      <td>Wurstküche</td>\n",
       "      <td>Arts District</td>\n",
       "      <td>4.0</td>\n",
       "      <td>/biz/wurstk%C3%BCche-los-angeles-2</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 687-4444</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>800 E 3rd St</td>\n",
       "      <td>['Hot Dogs' 'German' 'Gastropubs']</td>\n",
       "      <td>33.896347</td>\n",
       "      <td>-118.117083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>3</th>\n",
       "      <td>8644</td>\n",
       "      <td>False</td>\n",
       "      <td>Daikokuya Little Tokyo</td>\n",
       "      <td>Little Tokyo</td>\n",
       "      <td>4.0</td>\n",
       "      <td>/biz/daikokuya-little-tokyo-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 626-1680</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>327 E 1st St</td>\n",
       "      <td>['Ramen' 'Noodles']</td>\n",
       "      <td>34.049971</td>\n",
       "      <td>-118.240083</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>4</th>\n",
       "      <td>4641</td>\n",
       "      <td>False</td>\n",
       "      <td>Slurpin’ Ramen Bar - Los Angeles</td>\n",
       "      <td>Koreatown</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/slurpin-ramen-bar-los-angeles-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 388-8607</td>\n",
       "      <td>$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>3500 W 8th St</td>\n",
       "      <td>['Ramen' 'Noodles']</td>\n",
       "      <td>34.057590</td>\n",
       "      <td>-118.306725</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>5</th>\n",
       "      <td>6228</td>\n",
       "      <td>False</td>\n",
       "      <td>Bestia</td>\n",
       "      <td>Downtown</td>\n",
       "      <td>4.5</td>\n",
       "      <td>/biz/bestia-los-angeles</td>\n",
       "      <td>False</td>\n",
       "      <td>(213) 514-5724</td>\n",
       "      <td>$$$</td>\n",
       "      <td>[]</td>\n",
       "      <td>2121 E 7th Pl</td>\n",
       "      <td>['Italian' 'Cocktail Bars' 'Pizza']</td>\n",
       "      <td>34.033738</td>\n",
       "      <td>-118.229309</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "   reviewCount  renderAdInfo                              name  neighborhoods  \\\n",
       "1         6636         False                     Howlin’ Ray’s      Chinatown   \n",
       "2         8449         False                        Wurstküche  Arts District   \n",
       "3         8644         False            Daikokuya Little Tokyo   Little Tokyo   \n",
       "4         4641         False  Slurpin’ Ramen Bar - Los Angeles      Koreatown   \n",
       "5         6228         False                            Bestia       Downtown   \n",
       "\n",
       "   rating                                     businessUrl   isAd  \\\n",
       "1     4.5                  /biz/howlin-rays-los-angeles-3  False   \n",
       "2     4.0              /biz/wurstk%C3%BCche-los-angeles-2  False   \n",
       "3     4.0         /biz/daikokuya-little-tokyo-los-angeles  False   \n",
       "4     4.5  /biz/slurpin-ramen-bar-los-angeles-los-angeles  False   \n",
       "5     4.5                         /biz/bestia-los-angeles  False   \n",
       "\n",
       "            phone priceRange alternateNames formattedAddress  \\\n",
       "1  (213) 935-8399         $$             []   727 N Broadway   \n",
       "2  (213) 687-4444         $$             []     800 E 3rd St   \n",
       "3  (213) 626-1680         $$             []     327 E 1st St   \n",
       "4  (213) 388-8607         $$             []    3500 W 8th St   \n",
       "5  (213) 514-5724        $$$             []    2121 E 7th Pl   \n",
       "\n",
       "                                          categories   latitude   longitude  \n",
       "1  ['Southern' 'Chicken Shop' 'American (Traditio...  34.061519 -118.239473  \n",
       "2                 ['Hot Dogs' 'German' 'Gastropubs']  33.896347 -118.117083  \n",
       "3                                ['Ramen' 'Noodles']  34.049971 -118.240083  \n",
       "4                                ['Ramen' 'Noodles']  34.057590 -118.306725  \n",
       "5                ['Italian' 'Cocktail Bars' 'Pizza']  34.033738 -118.229309  "
      ]
     },
     "execution_count": 32,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 19.7 ms (started: 2021-03-10 08:02:22 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 32;\n",
       "                var nbb_unformatted_code = \"# Load the cleaned yelp dataset\\nyelp_filtered = pd.read_csv(\\\"./data/yelp_cleaned.csv\\\", index_col=[0])\\nyelp_filtered.head()\";\n",
       "                var nbb_formatted_code = \"# Load the cleaned yelp dataset\\nyelp_filtered = pd.read_csv(\\\"./data/yelp_cleaned.csv\\\", index_col=[0])\\nyelp_filtered.head()\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "# Load the cleaned yelp dataset\n",
    "yelp_filtered = pd.read_csv(\"./data/yelp_cleaned.csv\", index_col=[0])\n",
    "yelp_filtered.head()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "organized-investigator",
   "metadata": {},
   "source": [
    "### Problem Setup"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "buried-cooperative",
   "metadata": {},
   "source": [
    "Let $k \\in K$ index the users whose preference is to be accounted for.\n",
    "\n",
    "Let $m \\in M$ index the month we will be requesting a restaurant recommendation for.\n",
    "\n",
    "Let $n \\in N$ index the total number of restaurants in Los Angeles available on Yelp!.\n",
    "\n",
    "Let $\\mathcal{X} = \\left\\{{x}_{1}, {x}_{2}, \\cdots, {x}_{N}\\right\\}$ be a matrix that denotes our dataset of webscraped restaurant information from Yelp!, where"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "pregnant-chambers",
   "metadata": {},
   "source": [
    "\\begin{aligned}\n",
    "    {x}_{n} &= \n",
    "    \\begin{bmatrix}\n",
    "        \\text{name} = \\left\\{ \\mathbf{\\text{String}} \\right\\} \\\\\n",
    "        \\text{address} = \\left\\{ \\mathbf{\\text{String}}\\right\\} \\\\\n",
    "        \\text{neighbourhood} = \\left\\{ \\mathbf{\\text{String}} \\right\\} \\\\\n",
    "        \\text{num_reviews} = \\left\\{0.0 \\leq \\mathbf{\\text{Integer}} \\leq \\infty \\right\\} \\\\\n",
    "        \\text{rating} = \\left\\{ 0.0 \\leq \\mathbf{\\text{Float}} \\leq 5.0 \\right\\} \\\\\n",
    "        \\text{price_range} = \\left\\{ \\text{\\$}, \\text{\\$\\$}, \\text{\\$\\$\\$}, \\text{\\$\\$\\$\\$} \\right\\} \\\\\n",
    "        \\text{categories} = \\left\\{ \\text{Breakfast & Brunch}, \\text{Korean}, \\cdots, \\text{Halal} \\right\\} \\\\\n",
    "    \\end{bmatrix} \\\\\n",
    "\\end{aligned}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "wanted-henry",
   "metadata": {},
   "source": [
    "Let $\\mathcal{U} = \\left\\{{u}_{1}, {u}_{2}, \\cdots, {u}_{K}\\right\\}$ be a matrix, where"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "political-aside",
   "metadata": {},
   "source": [
    "\\begin{aligned}\n",
    "    {u}_{k} &= \n",
    "    \\begin{bmatrix}\n",
    "        \\text{address} = \\left\\{ \\mathbf{\\text{String}} \\right\\} \\\\\n",
    "        \\text{neighbourhood} = \\left\\{ \\mathbf{\\text{String}} \\right\\} \\\\\n",
    "        \\text{min_reviews} = \\left\\{0.0 \\leq \\mathbf{\\text{Integer}} \\leq \\infty \\right\\} \\\\\n",
    "        \\text{min_rating} = \\left\\{ 0.0 \\leq \\mathbf{\\text{Float}} \\leq 5.0 \\right\\} \\\\\n",
    "        \\text{price_range} = \\left\\{ \\text{\\$}, \\text{\\$\\$}, \\text{\\$\\$\\$}, \\text{\\$\\$\\$\\$} \\right\\} \\\\\n",
    "        \\text{max_distance (miles)} = \\left\\{ 0.0 \\leq \\bf{\\text{Float}} \\leq \\infty \\right\\} \\\\\n",
    "        \\text{categories} = \\left\\{ \\text{Breakfast & Brunch}, \\text{Korean}, \\cdots, \\text{Halal} \\right\\} \\\\\n",
    "    \\end{bmatrix} \\\\\n",
    "\\end{aligned}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "casual-diabetes",
   "metadata": {},
   "source": [
    "Let $w \\in \\mathbb{Z}^N_2 = \\left\\{0, 1\\right\\}^N$ a binary indicator vector denoting which of the $N$ restaurants in the Los Angeles area were chosen.\n",
    "\n",
    "Let $d: L \\times L \\rightarrow \\mathbb{R}^+_0, L = \\left\\{\\mathbf{\\text{Valid Address String}}\\right\\}$ be a function that calculates the distance in miles between two points.\n",
    "\n",
    "Let $A.\\text{<attribute>}$ denote the column vector (if $A$ is a matrix) / scalar (if $A$ is a vector) of just the specific attribute."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "advance-librarian",
   "metadata": {},
   "source": [
    "### Mixed-Integer Program Formulation"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "blank-popularity",
   "metadata": {},
   "source": [
    "\\begin{aligned}\n",
    "    \\underset{w}{\\text{maximize }} &{w^\\top}{\\left(\\mathcal{X}.\\text{rating}\\right)}  \\\\\n",
    "    \\text{subject to }\n",
    "    &w^\\top\\mathbb{1} = M \\\\\n",
    "    &w_nu_k.{\\text{min_reviews}} \\leq x_n.{\\text{num_reviews}}\\,\\forall\\,n \\in N, k \\in K \\\\\n",
    "    &w_nu_k.{\\text{min_rating}} \\leq x_n.{\\text{rating}}\\,\\forall\\,n \\in N, k \\in K \\\\\n",
    "    &w_n\\left(1 - \\frac{\\min_{n, k} d({{x_n}.{\\text{address}}}, {u_k}.{\\text{address}})}{\\max_{n, k} d({{x_n}.{\\text{address}}}, {u_k}.{\\text{address}})}\\right) \\leq \\alpha_n\\,\\forall\\,n \\in N, k \\in K, w_n = 1 \\\\\n",
    "    &u_k.{\\text{price_range}} \\geq x_n.{\\text{price_range}}\\,\\forall\\,n \\in N, k \\in K, w_n = 1 \\\\\n",
    "    &u_k.{\\text{neighbourhood}} = x_n.{\\text{neighbourhood}}\\,\\forall\\,n \\in N, k \\in K, w_n = 1 \\\\\n",
    "    &u_k.{\\text{categories}} \\subseteq x_n.{\\text{categories}}\\,\\forall\\,n \\in N, k \\in K, w_n = 1 \\\\\n",
    "\\end{aligned}"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "id": "macro-professional",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 476 ms (started: 2021-03-09 15:59:55 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 12;\n",
       "                var nbb_unformatted_code = \"# For distance calculation\\nfrom haversine import haversine, Unit\\n\\n# For optimization\\nimport cvxpy as cp\";\n",
       "                var nbb_formatted_code = \"# For distance calculation\\nfrom haversine import haversine, Unit\\n\\n# For optimization\\nimport cvxpy as cp\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "# For distance calculation\n",
    "from haversine import haversine, Unit\n",
    "\n",
    "# For optimization\n",
    "import cvxpy as cp\n",
    "\n",
    "# For converting street in LA to lat long\n",
    "locator = Nominatim(user_agent=\"myGeocoder\")\n",
    "\n",
    "def get_latlong(address: str, locator=locator):\n",
    "    \"\"\"Get lat, long from string address\"\"\"\n",
    "    location = locator.geocode(f\"{address}, Los Angeles, CA\")\n",
    "    try:\n",
    "        lat, long = location.latitude, location.longitude\n",
    "    except:\n",
    "        lat, long = None, None\n",
    "    return lat, long"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "entitled-factory",
   "metadata": {},
   "source": [
    "User Preference Matrix $\\mathcal{U}$:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "id": "fitting-attribute",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 2.64 s (started: 2021-03-09 17:18:06 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 21;\n",
       "                var nbb_unformatted_code = \"U = pd.DataFrame(\\n    [\\n        [\\n            \\\"3584 S Figueroa St\\\",\\n            None,\\n            100,\\n            4,\\n            \\\"$$\\\",\\n            None,\\n            None,\\n            *get_latlong(\\\"3584 S Figueroa St\\\"),\\n        ],  # Icon Plaza USC\\n        [\\n            \\\"3301 S Hoover St\\\",\\n            None,\\n            200,\\n            3.6,\\n            \\\"$$$\\\",\\n            None,\\n            None,\\n            *get_latlong(\\\"3301 S Hoover St\\\"),\\n        ],  # USC Village\\n        [\\n            \\\"10250 Santa Monica Blvd\\\",\\n            None,\\n            1000,\\n            4.5,\\n            \\\"$$$$\\\",\\n            None,\\n            None,\\n            *get_latlong(\\\"10250 Santa Monica Blvd\\\"),\\n        ],  # Westfield Century City\\n        [\\n            \\\"189 The Grove Dr\\\",\\n            None,\\n            1000,\\n            3.9,\\n            \\\"$$$\\\",\\n            None,\\n            [\\\"Korean\\\"],\\n            *get_latlong(\\\"189 The Grove Dr\\\"),\\n        ],  # The Grove\\n    ],\\n    columns=[\\n        \\\"address\\\",\\n        \\\"neighbourhood\\\",\\n        \\\"min_reviews\\\",\\n        \\\"min_rating\\\",\\n        \\\"price_range\\\",\\n        \\\"max_distance\\\",\\n        \\\"categories\\\",\\n        \\\"latitude\\\",\\n        \\\"longitude\\\",\\n    ],\\n)\\n\\nassert U[\\\"neighbourhood\\\"].nunique() <= 1, print(\\n    \\\"Number of different neighbourhood preferences must be <= 1.\\\"\\n)\\n# assert U[\\\"price_range\\\"].nunique() <= 1, print(\\n#     \\\"Number of different price_range preferences must be <= 1.\\\"\\n# )\\nassert (\\n    len(\\n        np.unique(\\n            [\\n                cat\\n                for cats in U[\\\"categories\\\"].to_numpy()\\n                if cats is not None\\n                for cat in cats\\n            ]\\n        )\\n    )\\n    <= 2\\n), print(\\\"Number of different category preferences must be <= 2.\\\")\";\n",
       "                var nbb_formatted_code = \"U = pd.DataFrame(\\n    [\\n        [\\n            \\\"3584 S Figueroa St\\\",\\n            None,\\n            100,\\n            4,\\n            \\\"$$\\\",\\n            None,\\n            None,\\n            *get_latlong(\\\"3584 S Figueroa St\\\"),\\n        ],  # Icon Plaza USC\\n        [\\n            \\\"3301 S Hoover St\\\",\\n            None,\\n            200,\\n            3.6,\\n            \\\"$$$\\\",\\n            None,\\n            None,\\n            *get_latlong(\\\"3301 S Hoover St\\\"),\\n        ],  # USC Village\\n        [\\n            \\\"10250 Santa Monica Blvd\\\",\\n            None,\\n            1000,\\n            4.5,\\n            \\\"$$$$\\\",\\n            None,\\n            None,\\n            *get_latlong(\\\"10250 Santa Monica Blvd\\\"),\\n        ],  # Westfield Century City\\n        [\\n            \\\"189 The Grove Dr\\\",\\n            None,\\n            1000,\\n            3.9,\\n            \\\"$$$\\\",\\n            None,\\n            [\\\"Korean\\\"],\\n            *get_latlong(\\\"189 The Grove Dr\\\"),\\n        ],  # The Grove\\n    ],\\n    columns=[\\n        \\\"address\\\",\\n        \\\"neighbourhood\\\",\\n        \\\"min_reviews\\\",\\n        \\\"min_rating\\\",\\n        \\\"price_range\\\",\\n        \\\"max_distance\\\",\\n        \\\"categories\\\",\\n        \\\"latitude\\\",\\n        \\\"longitude\\\",\\n    ],\\n)\\n\\nassert U[\\\"neighbourhood\\\"].nunique() <= 1, print(\\n    \\\"Number of different neighbourhood preferences must be <= 1.\\\"\\n)\\n# assert U[\\\"price_range\\\"].nunique() <= 1, print(\\n#     \\\"Number of different price_range preferences must be <= 1.\\\"\\n# )\\nassert (\\n    len(\\n        np.unique(\\n            [\\n                cat\\n                for cats in U[\\\"categories\\\"].to_numpy()\\n                if cats is not None\\n                for cat in cats\\n            ]\\n        )\\n    )\\n    <= 2\\n), print(\\\"Number of different category preferences must be <= 2.\\\")\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "U = pd.DataFrame(\n",
    "    [\n",
    "        [\n",
    "            \"3584 S Figueroa St\",\n",
    "            None,\n",
    "            100,\n",
    "            4,\n",
    "            \"$$\",\n",
    "            None,\n",
    "            None,\n",
    "            *get_latlong(\"3584 S Figueroa St\"),\n",
    "        ],  # Icon Plaza USC\n",
    "        [\n",
    "            \"3301 S Hoover St\",\n",
    "            None,\n",
    "            200,\n",
    "            3.6,\n",
    "            \"$$$\",\n",
    "            None,\n",
    "            None,\n",
    "            *get_latlong(\"3301 S Hoover St\"),\n",
    "        ],  # USC Village\n",
    "        [\n",
    "            \"10250 Santa Monica Blvd\",\n",
    "            None,\n",
    "            1000,\n",
    "            4.5,\n",
    "            \"$$$$\",\n",
    "            None,\n",
    "            None,\n",
    "            *get_latlong(\"10250 Santa Monica Blvd\"),\n",
    "        ],  # Westfield Century City\n",
    "        [\n",
    "            \"189 The Grove Dr\",\n",
    "            None,\n",
    "            1000,\n",
    "            3.9,\n",
    "            \"$$$\",\n",
    "            None,\n",
    "            [\"Korean\"],\n",
    "            *get_latlong(\"189 The Grove Dr\"),\n",
    "        ],  # The Grove\n",
    "    ],\n",
    "    columns=[\n",
    "        \"address\",\n",
    "        \"neighbourhood\",\n",
    "        \"min_reviews\",\n",
    "        \"min_rating\",\n",
    "        \"price_range\",\n",
    "        \"max_distance\",\n",
    "        \"categories\",\n",
    "        \"latitude\",\n",
    "        \"longitude\",\n",
    "    ],\n",
    ")\n",
    "\n",
    "assert U[\"neighbourhood\"].nunique() <= 1, print(\n",
    "    \"Number of different neighbourhood preferences must be <= 1.\"\n",
    ")\n",
    "# assert U[\"price_range\"].nunique() <= 1, print(\n",
    "#     \"Number of different price_range preferences must be <= 1.\"\n",
    "# )\n",
    "assert (\n",
    "    len(\n",
    "        np.unique(\n",
    "            [\n",
    "                cat\n",
    "                for cats in U[\"categories\"].to_numpy()\n",
    "                if cats is not None\n",
    "                for cat in cats\n",
    "            ]\n",
    "        )\n",
    "    )\n",
    "    <= 2\n",
    "), print(\"Number of different category preferences must be <= 2.\")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "forced-cooling",
   "metadata": {},
   "source": [
    "Updated Data Matrix $\\mathcal{X}$, filtering out some data that does not match constraints:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "id": "anonymous-welcome",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 15.4 ms (started: 2021-03-10 07:57:39 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 27;\n",
       "                var nbb_unformatted_code = \"X = yelp_filtered[\\n    [\\n        \\\"name\\\",\\n        \\\"formattedAddress\\\",\\n        \\\"neighborhoods\\\",\\n        \\\"reviewCount\\\",\\n        \\\"rating\\\",\\n        \\\"priceRange\\\",\\n        \\\"categories\\\",\\n        \\\"latitude\\\",\\n        \\\"longitude\\\",\\n    ]\\n]\\nX.columns = [\\n    \\\"name\\\",\\n    \\\"address\\\",\\n    \\\"neighbourhood\\\",\\n    \\\"num_reviews\\\",\\n    \\\"rating\\\",\\n    \\\"price_range\\\",\\n    \\\"categories\\\",\\n    \\\"latitude\\\",\\n    \\\"longitude\\\",\\n]  # rename columns\\n\\n# Filter out the neighbourhoods not in the user preference\\nif len(U[\\\"neighbourhood\\\"].dropna().unique()) > 0:\\n    X = X[X[\\\"neighbourhood\\\"] == U[\\\"neighbourhood\\\"].dropna().unique()[0]]\\n\\n# Filter out the restaurants whose price ranges exceed the lowest price range in user preferences\\nif len(U[\\\"price_range\\\"].dropna().unique()) > 0:\\n    X = X[X[\\\"price_range\\\"] == sorted(U[\\\"price_range\\\"].dropna().unique(), key=lambda x: len(x), reverse=False)[0]]\\n\\n# Filter out restaurants that are not in the same categories as what we requested in user preference\\nif len(U[\\\"categories\\\"].dropna()) > 0:\\n    X = X[\\n        X[\\\"categories\\\"].apply(\\n            lambda categories: np.any(\\n                [\\n                    cat in categories\\n                    for cat in [\\n                        cat\\n                        for cats in U[\\\"categories\\\"].to_numpy()\\n                        if cats is not None\\n                        for cat in cats\\n                    ]\\n                ]\\n            )\\n        )\\n    ]\";\n",
       "                var nbb_formatted_code = \"X = yelp_filtered[\\n    [\\n        \\\"name\\\",\\n        \\\"formattedAddress\\\",\\n        \\\"neighborhoods\\\",\\n        \\\"reviewCount\\\",\\n        \\\"rating\\\",\\n        \\\"priceRange\\\",\\n        \\\"categories\\\",\\n        \\\"latitude\\\",\\n        \\\"longitude\\\",\\n    ]\\n]\\nX.columns = [\\n    \\\"name\\\",\\n    \\\"address\\\",\\n    \\\"neighbourhood\\\",\\n    \\\"num_reviews\\\",\\n    \\\"rating\\\",\\n    \\\"price_range\\\",\\n    \\\"categories\\\",\\n    \\\"latitude\\\",\\n    \\\"longitude\\\",\\n]  # rename columns\\n\\n# Filter out the neighbourhoods not in the user preference\\nif len(U[\\\"neighbourhood\\\"].dropna().unique()) > 0:\\n    X = X[X[\\\"neighbourhood\\\"] == U[\\\"neighbourhood\\\"].dropna().unique()[0]]\\n\\n# Filter out the restaurants whose price ranges exceed the lowest price range in user preferences\\nif len(U[\\\"price_range\\\"].dropna().unique()) > 0:\\n    X = X[\\n        X[\\\"price_range\\\"]\\n        == sorted(\\n            U[\\\"price_range\\\"].dropna().unique(), key=lambda x: len(x), reverse=False\\n        )[0]\\n    ]\\n\\n# Filter out restaurants that are not in the same categories as what we requested in user preference\\nif len(U[\\\"categories\\\"].dropna()) > 0:\\n    X = X[\\n        X[\\\"categories\\\"].apply(\\n            lambda categories: np.any(\\n                [\\n                    cat in categories\\n                    for cat in [\\n                        cat\\n                        for cats in U[\\\"categories\\\"].to_numpy()\\n                        if cats is not None\\n                        for cat in cats\\n                    ]\\n                ]\\n            )\\n        )\\n    ]\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "X = yelp_filtered[\n",
    "    [\n",
    "        \"name\",\n",
    "        \"formattedAddress\",\n",
    "        \"neighborhoods\",\n",
    "        \"reviewCount\",\n",
    "        \"rating\",\n",
    "        \"priceRange\",\n",
    "        \"categories\",\n",
    "        \"latitude\",\n",
    "        \"longitude\",\n",
    "    ]\n",
    "]\n",
    "X.columns = [\n",
    "    \"name\",\n",
    "    \"address\",\n",
    "    \"neighbourhood\",\n",
    "    \"num_reviews\",\n",
    "    \"rating\",\n",
    "    \"price_range\",\n",
    "    \"categories\",\n",
    "    \"latitude\",\n",
    "    \"longitude\",\n",
    "]  # rename columns\n",
    "\n",
    "# Filter out the neighbourhoods not in the user preference\n",
    "if len(U[\"neighbourhood\"].dropna().unique()) > 0:\n",
    "    X = X[X[\"neighbourhood\"] == U[\"neighbourhood\"].dropna().unique()[0]]\n",
    "\n",
    "# Filter out the restaurants whose price ranges exceed the lowest price range in user preferences\n",
    "if len(U[\"price_range\"].dropna().unique()) > 0:\n",
    "    X = X[\n",
    "        X[\"price_range\"]\n",
    "        == sorted(\n",
    "            U[\"price_range\"].dropna().unique(), key=lambda x: len(x), reverse=False\n",
    "        )[0]\n",
    "    ]\n",
    "\n",
    "# Filter out restaurants that are not in the same categories as what we requested in user preference\n",
    "if len(U[\"categories\"].dropna()) > 0:\n",
    "    X = X[\n",
    "        X[\"categories\"].apply(\n",
    "            lambda categories: np.any(\n",
    "                [\n",
    "                    cat in categories\n",
    "                    for cat in [\n",
    "                        cat\n",
    "                        for cats in U[\"categories\"].to_numpy()\n",
    "                        if cats is not None\n",
    "                        for cat in cats\n",
    "                    ]\n",
    "                ]\n",
    "            )\n",
    "        )\n",
    "    ]"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "pressing-albania",
   "metadata": {},
   "source": [
    "Optimization"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 29,
   "id": "victorian-queens",
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Mixed Integer Programming Solution\n",
      "==============================\n",
      "Status: optimal\n",
      "The optimal value is: 22.5\n",
      "Restaurants chosen: \n"
     ]
    },
    {
     "data": {
      "text/html": [
       "<div>\n",
       "<style scoped>\n",
       "    .dataframe tbody tr th:only-of-type {\n",
       "        vertical-align: middle;\n",
       "    }\n",
       "\n",
       "    .dataframe tbody tr th {\n",
       "        vertical-align: top;\n",
       "    }\n",
       "\n",
       "    .dataframe thead th {\n",
       "        text-align: right;\n",
       "    }\n",
       "</style>\n",
       "<table border=\"1\" class=\"dataframe\">\n",
       "  <thead>\n",
       "    <tr style=\"text-align: right;\">\n",
       "      <th></th>\n",
       "      <th>name</th>\n",
       "      <th>address</th>\n",
       "      <th>neighbourhood</th>\n",
       "      <th>num_reviews</th>\n",
       "      <th>rating</th>\n",
       "      <th>price_range</th>\n",
       "      <th>categories</th>\n",
       "      <th>latitude</th>\n",
       "      <th>longitude</th>\n",
       "    </tr>\n",
       "  </thead>\n",
       "  <tbody>\n",
       "    <tr>\n",
       "      <th>21</th>\n",
       "      <td>Yup Dduk LA</td>\n",
       "      <td>3603 W 6th St</td>\n",
       "      <td>Wilshire Center</td>\n",
       "      <td>2111</td>\n",
       "      <td>4.5</td>\n",
       "      <td>$$</td>\n",
       "      <td>[Korean, Chicken Shop]</td>\n",
       "      <td>34.063892</td>\n",
       "      <td>-118.300805</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>32</th>\n",
       "      <td>Han Bat Sul Lung Tang</td>\n",
       "      <td>4163 W 5th St</td>\n",
       "      <td>Koreatown</td>\n",
       "      <td>2294</td>\n",
       "      <td>4.5</td>\n",
       "      <td>$$</td>\n",
       "      <td>[Korean, Comfort Food, Soup]</td>\n",
       "      <td>34.065408</td>\n",
       "      <td>-118.309849</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>36</th>\n",
       "      <td>Magal BBQ</td>\n",
       "      <td>3460 W 8th St</td>\n",
       "      <td>Koreatown</td>\n",
       "      <td>1676</td>\n",
       "      <td>4.5</td>\n",
       "      <td>$$</td>\n",
       "      <td>[Korean, Barbeque]</td>\n",
       "      <td>34.057598</td>\n",
       "      <td>-118.305479</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>50</th>\n",
       "      <td>Bulgogi Hut</td>\n",
       "      <td>3600 Wilshire Blvd</td>\n",
       "      <td>Koreatown</td>\n",
       "      <td>2471</td>\n",
       "      <td>4.5</td>\n",
       "      <td>$$</td>\n",
       "      <td>[Korean, Barbeque, Asian Fusion]</td>\n",
       "      <td>34.062375</td>\n",
       "      <td>-118.298589</td>\n",
       "    </tr>\n",
       "    <tr>\n",
       "      <th>68</th>\n",
       "      <td>Eight Korean BBQ</td>\n",
       "      <td>863 S Western Ave</td>\n",
       "      <td>Koreatown</td>\n",
       "      <td>1651</td>\n",
       "      <td>4.5</td>\n",
       "      <td>$$</td>\n",
       "      <td>[Korean, Barbeque]</td>\n",
       "      <td>34.056027</td>\n",
       "      <td>-118.309888</td>\n",
       "    </tr>\n",
       "  </tbody>\n",
       "</table>\n",
       "</div>"
      ],
      "text/plain": [
       "                     name             address    neighbourhood  num_reviews  \\\n",
       "21            Yup Dduk LA       3603 W 6th St  Wilshire Center         2111   \n",
       "32  Han Bat Sul Lung Tang       4163 W 5th St        Koreatown         2294   \n",
       "36              Magal BBQ       3460 W 8th St        Koreatown         1676   \n",
       "50            Bulgogi Hut  3600 Wilshire Blvd        Koreatown         2471   \n",
       "68       Eight Korean BBQ   863 S Western Ave        Koreatown         1651   \n",
       "\n",
       "    rating price_range                        categories   latitude  \\\n",
       "21     4.5          $$            [Korean, Chicken Shop]  34.063892   \n",
       "32     4.5          $$      [Korean, Comfort Food, Soup]  34.065408   \n",
       "36     4.5          $$                [Korean, Barbeque]  34.057598   \n",
       "50     4.5          $$  [Korean, Barbeque, Asian Fusion]  34.062375   \n",
       "68     4.5          $$                [Korean, Barbeque]  34.056027   \n",
       "\n",
       "     longitude  \n",
       "21 -118.300805  \n",
       "32 -118.309849  \n",
       "36 -118.305479  \n",
       "50 -118.298589  \n",
       "68 -118.309888  "
      ]
     },
     "execution_count": 29,
     "metadata": {},
     "output_type": "execute_result"
    },
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "time: 283 ms (started: 2021-03-10 07:59:21 +08:00)\n"
     ]
    },
    {
     "data": {
      "application/javascript": [
       "\n",
       "            setTimeout(function() {\n",
       "                var nbb_cell_id = 29;\n",
       "                var nbb_unformatted_code = \"# Distance metric\\nd = lambda lat1, long1, lat2, long2: haversine(\\n    (lat1, long1), (lat2, long2), unit=Unit.MILES\\n)\\n\\n# Percentage difference between furthest travelling indiividual and shortest travelling individual\\n\\u03b1 = 0.75\\n\\n# Number of months we getting recommendations\\nM = 5\\n\\n# Number of restaurants in our dataset\\nN = X.shape[0]\\n\\n# Number of users \\nK = U.shape[0]\\n\\n# Create one vector optimization variable.\\nw = cp.Variable(X.shape[0], boolean=True)\\n\\n# Create constraints.\\nconstraints = [\\n    cp.sum(w) >= M,\\n    cp.sum(w) <= M,\\n    *[\\n        w_n * u_k <= x_n\\n        for u_k in U[\\\"min_reviews\\\"]\\n        for w_n, x_n in zip(w, X[\\\"num_reviews\\\"])\\n    ],\\n    *[w_n * u_k <= x_n for u_k in U[\\\"min_rating\\\"] for w_n, x_n in zip(w, X[\\\"rating\\\"])],\\n    *[\\n        w_n\\n        * (\\n            1\\n            - (\\n                cp.minimum(\\n                    *[\\n                        d(row[\\\"latitude\\\"], row[\\\"longitude\\\"], lat_n, long_n)\\n                        for idx, row in U[[\\\"latitude\\\", \\\"longitude\\\"]].iterrows()\\n                    ]\\n                )\\n                / cp.maximum(\\n                    *[\\n                        d(row[\\\"latitude\\\"], row[\\\"longitude\\\"], lat_n, long_n)\\n                        for idx, row in U[[\\\"latitude\\\", \\\"longitude\\\"]].iterrows()\\n                    ]\\n                )\\n            )\\n        )\\n        <= \\u03b1\\n        for w_n, lat_n, long_n in zip(w, X[\\\"latitude\\\"], X[\\\"longitude\\\"])\\n    ],\\n]\\n\\n# Form objective.\\nobj = cp.Maximize(w.T @ X[\\\"rating\\\"])\\n\\n# Form and solve problem.\\nprob = cp.Problem(obj, constraints)\\nprob.solve()\\n\\nprint(\\\"Mixed Integer Programming Solution\\\")\\nprint(\\\"=\\\" * 30)\\nprint(f\\\"Status: {prob.status}\\\")\\nprint(f\\\"The optimal value is: {np.round(prob.value, 2)}\\\")\\nprint(\\\"Restaurants chosen: \\\")\\nX.iloc[np.argwhere(w.value).flatten()]\";\n",
       "                var nbb_formatted_code = \"# Distance metric\\nd = lambda lat1, long1, lat2, long2: haversine(\\n    (lat1, long1), (lat2, long2), unit=Unit.MILES\\n)\\n\\n# Percentage difference between furthest travelling indiividual and shortest travelling individual\\n\\u03b1 = 0.75\\n\\n# Number of months we getting recommendations\\nM = 5\\n\\n# Number of restaurants in our dataset\\nN = X.shape[0]\\n\\n# Number of users\\nK = U.shape[0]\\n\\n# Create one vector optimization variable.\\nw = cp.Variable(X.shape[0], boolean=True)\\n\\n# Create constraints.\\nconstraints = [\\n    cp.sum(w) >= M,\\n    cp.sum(w) <= M,\\n    *[\\n        w_n * u_k <= x_n\\n        for u_k in U[\\\"min_reviews\\\"]\\n        for w_n, x_n in zip(w, X[\\\"num_reviews\\\"])\\n    ],\\n    *[w_n * u_k <= x_n for u_k in U[\\\"min_rating\\\"] for w_n, x_n in zip(w, X[\\\"rating\\\"])],\\n    *[\\n        w_n\\n        * (\\n            1\\n            - (\\n                cp.minimum(\\n                    *[\\n                        d(row[\\\"latitude\\\"], row[\\\"longitude\\\"], lat_n, long_n)\\n                        for idx, row in U[[\\\"latitude\\\", \\\"longitude\\\"]].iterrows()\\n                    ]\\n                )\\n                / cp.maximum(\\n                    *[\\n                        d(row[\\\"latitude\\\"], row[\\\"longitude\\\"], lat_n, long_n)\\n                        for idx, row in U[[\\\"latitude\\\", \\\"longitude\\\"]].iterrows()\\n                    ]\\n                )\\n            )\\n        )\\n        <= \\u03b1\\n        for w_n, lat_n, long_n in zip(w, X[\\\"latitude\\\"], X[\\\"longitude\\\"])\\n    ],\\n]\\n\\n# Form objective.\\nobj = cp.Maximize(w.T @ X[\\\"rating\\\"])\\n\\n# Form and solve problem.\\nprob = cp.Problem(obj, constraints)\\nprob.solve()\\n\\nprint(\\\"Mixed Integer Programming Solution\\\")\\nprint(\\\"=\\\" * 30)\\nprint(f\\\"Status: {prob.status}\\\")\\nprint(f\\\"The optimal value is: {np.round(prob.value, 2)}\\\")\\nprint(\\\"Restaurants chosen: \\\")\\nX.iloc[np.argwhere(w.value).flatten()]\";\n",
       "                var nbb_cells = Jupyter.notebook.get_cells();\n",
       "                for (var i = 0; i < nbb_cells.length; ++i) {\n",
       "                    if (nbb_cells[i].input_prompt_number == nbb_cell_id) {\n",
       "                        if (nbb_cells[i].get_text() == nbb_unformatted_code) {\n",
       "                             nbb_cells[i].set_text(nbb_formatted_code);\n",
       "                        }\n",
       "                        break;\n",
       "                    }\n",
       "                }\n",
       "            }, 500);\n",
       "            "
      ],
      "text/plain": [
       "<IPython.core.display.Javascript object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "# Distance metric\n",
    "d = lambda lat1, long1, lat2, long2: haversine(\n",
    "    (lat1, long1), (lat2, long2), unit=Unit.MILES\n",
    ")\n",
    "\n",
    "# Percentage difference between furthest travelling indiividual and shortest travelling individual\n",
    "α = 0.75\n",
    "\n",
    "# Number of months we getting recommendations\n",
    "M = 5\n",
    "\n",
    "# Number of restaurants in our dataset\n",
    "N = X.shape[0]\n",
    "\n",
    "# Number of users\n",
    "K = U.shape[0]\n",
    "\n",
    "# Create one vector optimization variable.\n",
    "w = cp.Variable(X.shape[0], boolean=True)\n",
    "\n",
    "# Create constraints.\n",
    "constraints = [\n",
    "    cp.sum(w) >= M,\n",
    "    cp.sum(w) <= M,\n",
    "    *[\n",
    "        w_n * u_k <= x_n\n",
    "        for u_k in U[\"min_reviews\"]\n",
    "        for w_n, x_n in zip(w, X[\"num_reviews\"])\n",
    "    ],\n",
    "    *[w_n * u_k <= x_n for u_k in U[\"min_rating\"] for w_n, x_n in zip(w, X[\"rating\"])],\n",
    "    *[\n",
    "        w_n\n",
    "        * (\n",
    "            1\n",
    "            - (\n",
    "                cp.minimum(\n",
    "                    *[\n",
    "                        d(row[\"latitude\"], row[\"longitude\"], lat_n, long_n)\n",
    "                        for idx, row in U[[\"latitude\", \"longitude\"]].iterrows()\n",
    "                    ]\n",
    "                )\n",
    "                / cp.maximum(\n",
    "                    *[\n",
    "                        d(row[\"latitude\"], row[\"longitude\"], lat_n, long_n)\n",
    "                        for idx, row in U[[\"latitude\", \"longitude\"]].iterrows()\n",
    "                    ]\n",
    "                )\n",
    "            )\n",
    "        )\n",
    "        <= α\n",
    "        for w_n, lat_n, long_n in zip(w, X[\"latitude\"], X[\"longitude\"])\n",
    "    ],\n",
    "]\n",
    "\n",
    "# Form objective.\n",
    "obj = cp.Maximize(w.T @ X[\"rating\"])\n",
    "\n",
    "# Form and solve problem.\n",
    "prob = cp.Problem(obj, constraints)\n",
    "prob.solve()\n",
    "\n",
    "print(\"Mixed Integer Programming Solution\")\n",
    "print(\"=\" * 30)\n",
    "print(f\"Status: {prob.status}\")\n",
    "print(f\"The optimal value is: {np.round(prob.value, 2)}\")\n",
    "print(\"Restaurants chosen: \")\n",
    "X.iloc[np.argwhere(w.value).flatten()]"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.5"
  },
  "latex_envs": {
   "LaTeX_envs_menu_present": true,
   "autoclose": false,
   "autocomplete": true,
   "bibliofile": "biblio.bib",
   "cite_by": "apalike",
   "current_citInitial": 1,
   "eqLabelWithNumbers": true,
   "eqNumInitial": 1,
   "hotkeys": {
    "equation": "Ctrl-E",
    "itemize": "Ctrl-I"
   },
   "labels_anchors": false,
   "latex_user_defs": false,
   "report_style_numbering": false,
   "user_envs_cfg": false
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
